#include "math.h"
#include "resample.h"

/* periodic symmetric extension of src */
#define REFOLD(J,L)   { if(J<0) J =-J; J = J % (2*L-1); if(J>=L) J=(2*L-1)-J;}
/* border extension of src */
#define BORD(J,L)   { if(J<0) J=0; else if (J>L) J=L; }
/* zero extension of src */
#define ZERO(J,L)   { if(J<0 || J>L) continue; }

/* enforce boundary condition <-> array bounds */
//#define BC(J,L) REFOLD(J,L)



#undef  MAX
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))

#undef  MIN
#define MIN(a, b)  (((a) < (b)) ? (a) : (b))

#undef  ABS
#define ABS(a)     (((a) < 0) ? -(a) : (a))

#undef  CLAMP
#define CLAMP(x, low, high)  (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))


#include "stdio.h"
#include "math.h"


/* #define inline */
/* #define static */
/* #warning static and inline are disabled */

// WARNING this is not ANTIALIASED in any way but is very simple and fast

static inline
void
resample_array_inv_noaa_
  (const double *F, 
   const unsigned char *src, int s_len, int s_stride, 
         unsigned char *dst, int d_len, int d_stride)
{
  int i,p=0,j;
  for(i=0;i<d_len;i++) {
    j=F[i];
    BORD(j,s_len)
    dst[p]=src[j* s_stride];
    p+= d_stride;
  }
}


/*********** antialiased using bilinear interpixel**************
 * works well when the scale is not reduced and/or when there are not
 * very fine grains; it is much faster; it is approximate in many points
 * For animations it is better to use antialiased
 */

static
void
resample_array_inv_bilin
  (const double *F, 
   const unsigned char *src, int s_len, int s_stride, 
         unsigned char *dst, int d_len, int d_stride)
{
  int i,p=0,j,nj,oj;
  double v;
  oj=floor(F[0]);
  BORD(oj,s_len);
  for(i=0;i<d_len;i++) {
    v=0;
    j=floor(F[i]);
    if(j>=0 && j <s_len) {
#ifdef SEEMS_UNNEEDED
      int c;
      /* I have tested the code and this seems to introduce ripples;
      * I don't want to improve it, use AA if you really want AA*/
      if( oj>=j+2) {
	c=oj-j+1;
	if(oj>=s_len) oj=s_len-1;
	while (oj>=j) { v+=src[oj* s_stride]; oj--;  }
	v/=c;
      }else 
	if ( oj+2<=j) {
	  c=j-oj+1;
	  if(oj<0) oj=0;
	  while (oj<=j) { v+=src[oj* s_stride]; oj++;  }
	  v/=c;
	} else 
#endif
	  {      //if( ox-x<=1 && x-ox<=1) 
	    double dj=F[i]-j;
	    nj=j+1;
	    if(nj>=s_len) nj=s_len-1;
	    v=src[j* s_stride]*(1-dj) + src[nj* s_stride]*dj;
	  }
    }
    dst[p]=v;
    p+= d_stride;
    oj=j;
  }
}
/*********** antialiased using convolutional kernels***************/

#include "sinc_table.h"

static double sinc_real(const double x)
{
  if (x<0.001 && x > -0.001)
    return 1;
  else
   return sin(M_PI*x) / M_PI /x;
}

static double sinc_by_table(const double x)
{
  if (x < -4 || x > 4)
    return sin(M_PI*x) / M_PI /x;
  else
    return sinc_table [(int)((x+4)*100. )];
}


static double lanczos(const double x)
{
  if (x < -2 || x > 2)
    return 0;
  else
    return sinc_table [(int)((x+4)*100. )]; 
}

static double lanczos4(const double x)
{
  if (x < -4 || x > 4)
    return 0;
  else
    return sinc_table [(int)((x+4)*100. )]; 
}

static double triangle(const double x)
{
  if (x<-1) return 0;
  else if(x<0) return x+1;
  else if(x<1) return 1-x;
  else return 0;
}


static 
void
resample_array_inv_conv
  (const double *F, 
   const unsigned char *src, int s_len, int s_stride, 
         unsigned char *dst, int d_len, int d_stride)
{
  int i,p=0,j, lastj;
  double x,nx,px,c,s,v,dx;
  for(i=0;i<d_len;i++) {
    v=0; c=0;
    x=F[i];
    nx=F[i+1];
    px=F[i-1];
    dx=ABS(nx-px)/2.;
    if(dx<1) dx=1;
    lastj=ceil(x+dx+dx);
    for (j=floor(x-dx-dx) ; j<=lastj  ; j++ ) {
      s=lanczos( (x-(double)j) / dx );
      if ( j>=0 && j<s_len)
	v+=s* (double)src[j* s_stride];
      c+=s;
    }
    if(c<0.0001 && c > -0.0001)
      fprintf(stderr,"%s:%s:%d: too small c=%f\n",__FILE__,__FUNCTION__,__LINE__,c);
    else
      v/=c;
    dst[p]=CLAMP(v,0,255);
	
    p+= d_stride;    
  }
}



void (*resample_array_inv)(const double *F, 
   const unsigned char *src, int s_len, int s_stride, 
   unsigned char *dst, int d_len, int d_stride)= resample_array_inv_conv;



void
mesh_resample_use_aa(int f)
{
  if (f)
    resample_array_inv=resample_array_inv_conv;
  else
    resample_array_inv=resample_array_inv_bilin;
}
